(ns muneem.import
  (:require [clojure.data.csv :as csv]
            [clojure.java.io]
            [cljc.java-time.local-date :as jt]
            [cljc.java-time.format.date-time-formatter :as pt]
            [clojure.java.shell :refer [sh]]
            [clojure.string :as string]
            [clojure.string]))

(def beancount-file "./documents/ledger.beancount")

(def transaction-details (partial #(nth % 5)))
(def category (partial #(nth % 7)))
(def balance (partial #(nth % 6)))
(defn merchant-name [trans]
  (if (= (count trans) 9)
    (nth trans 8)
    ""))
(def amount (partial #(nth % 1)))
(def trans-date (partial #(nth % 0)))
(def transaction-type (partial #(nth % 4)))

(defn take-csv
  "Takes file name and reads data."
  [file]
  (-> (csv/read-csv (slurp file))
      ; We don't care about the header
      (rest)))

(defn make-date
  "Reads date from the 5th column (Transaction Details)
  The 2nd element split by space is DD/MM of the actal transaction
  Add 2022 to that and use it as the transaction date"
  [trans statement]
  ; TODO this doesn't work for various transactions and other years
  ; Parse real date -> check that it is in fact a date
  ; Use normal date to determine year
  ; If it's not a date, use regular date
  (let [real-date (-> trans
                      transaction-details
                      (clojure.string/split #" ")
                      second)
        trans-date-year (-> (trans-date trans)
                            (clojure.string/split #" ")
                            (last)
                            (->> (str "20")))
        with-year (-> real-date
                      (str "/" trans-date-year)
                      (clojure.string/split #"/")
                      reverse)
        parse-year (-> (trans-date trans)
                       (jt/parse (pt/of-pattern "dd MMM uu"))
                       (jt/format (pt/of-pattern "yyyy-MM-dd")))]
    (if (and (not (nil? real-date)) (re-matches #"../.." real-date))
      (conj statement (clojure.string/join "-" with-year))
      (conj statement parse-year))))

(defn add-transaction-flag
  "Bean count requires transaction type, we use * which means txn"
  [statement]
  (conj statement "*"))

(defn add-narration
  "Add some comments about the transaction from the 5th column (Transaction Details)"
  [trans statement]
  (let [narration (-> trans
                      transaction-details
                      (clojure.string/split #" ")
                      ; first 2 are payment method and date
                      rest
                      rest
                      (conj "\"")
                      (cons "\""))]
    (conj statement narration)))

(defn sanitize
  [string]
  (-> string
      (clojure.string/capitalize)
      (clojure.string/replace #" " "")
      (clojure.string/replace #"'" "")
      (clojure.string/replace #"\." "")
      (clojure.string/replace #"\(.+\)" "")
      (clojure.string/replace #"&" "-")))

; These should all be the same method tbh
; Or they should be a map?
; check out how-to-rules for more

(defn add-account-from-and-amount
  "We're only handling checking account for now
  All amounts are from it's perspsective"
  [trans statement]
  (let [amount (amount trans)
        account "Assets:Checking"
        string (str "  " account "     " amount " AUD")]
    (conj statement string)))

(defn add-account-from-and-amount-savings
  [trans statement]
  (let [amount (amount trans)
        account "Assets:Savings"
        string (str "  " account "     " amount " AUD")]
    (conj statement string)))

(defn debit-account-to
  [trans statement]
  (let [category (sanitize (category trans))
        merchant-name (sanitize (merchant-name trans))
                   ; Uncategorized strings need to be reconciled
                   ; TODO we'd probably want rules based on the transaction details
                   ; to automatically categorize further ones
        account-to (if (= category "Uncategorised")
                     (if (> (count merchant-name) 0)
                       (str "  ! Expenses:" category ":" merchant-name)
                       (str "  ! Expenses:" category))
                     (if (> (count merchant-name) 0)
                       (str "  Expenses:" category ":" merchant-name)
                       (str "  Expenses:" category)))]
    (conj statement account-to)))

(defn fees-account-to
  [trans statement]
  (let [category (sanitize (category trans))
        merchant-name (sanitize (merchant-name trans))
        account-to (str "  Expenses:Fees")]
    (conj statement account-to)))

(defn atm-debit-account-to
  [trans statement]
  (let [category (sanitize (category trans))
        merchant-name (sanitize (merchant-name trans))
        account-to (str "  Assets:Cash")]
    (conj statement account-to)))

(defn transfer-credit-account-to
  [trans statement]
  (let [account-to (str "  Assets:Savings")
        income (str "  ! Income:Other")
        details (transaction-details trans)
        category (category trans)]
    (println category)
    (if (= category "Transfers in")
      (conj statement income)
      (if (boolean (re-find #".*Linked Acc.*" details))
        (conj statement account-to)
        (conj statement income)))))

(defn transfer-debit-savings
  "This function is going to change a lot
  It currently only works with Expenses
  but in the future will need to handle determining account to"
  [trans statement]
  (let [details (transaction-details trans)
        account-to "  ! Expenses:Uncategorised:From-Savings"]
    (if (boolean (re-find #".*Linked Acc.*" details))
      nil ; these are the ones we want to skip, we accounted for them in checking
      (conj statement account-to))))

(defn bank-credit-account-to
  [_ statement]
  (let [account-to (str "  ! Income:ThreatDefence")]
    (conj statement account-to)))

(defn dividend-savings
  [_ statement]
  (let [account-to (str "  Income:Dividends")]
    (conj statement account-to)))

(defn salary-savings
  [_ statement]
  (let [account-to (str "  Income:Cubiko")]
    (conj statement account-to)))

(defn interest-account-to
  [_ statement]
  (let [account-to (str "  Expenses:NAB")]
    (conj statement account-to)))

(defn interest-savings
  [_ statement]
  (let [account-to (str "  Income:NAB")]
    (conj statement account-to)))

(defn automatic-drawing-account-to
  [_ statement]
  (let [account-to (str "  Assets:Spaceship")]
    (conj statement account-to)))

(defn misc-credit-account-to
  [_ statement]
  (let [account-to (str "  Income:Beem")]
    (conj statement account-to)))

(defn deposit-account-to
  "This function is going to change a lot
  It currently only works with Expenses
  but in the future will need to handle determining account to"
  [trans statement]
  (let [account-to (str "  ! Expenses:OverCharged")]
    (conj statement account-to)))

(defn transfer-debit-account-to
  "This function is going to change a lot
  It currently only works with Expenses
  but in the future will need to handle determining account to"
  [trans statement]
  (let [rent (str "  Expenses:Rent")
        trans-type (transaction-type trans)
        amt (amount trans)
        else-account (str "  Expenses:" (-> trans-type
                                            (clojure.string/split #" ")
                                            (nth 1)))]
    (if (or (= amt 170) (= amt 540))
      (conj statement rent)
      (conj statement else-account))))

(defn transfer-credit-savings
  "This function is going to change a lot
  It currently only works with Expenses
  but in the future will need to handle determining account to"
  [trans statement]
  (let [account-to (str "  Assets:Checking")]
    (conj statement account-to)))

;; RULES

(defn determine-category
  "We probably want some sort of rule engine 
  that takes rules and applies the first match or something (see docs)
  but this will do for now
  
  Writes out the account-to to the statement"
  [trans statement]
  (let [trans-type (transaction-type trans)]
    (case trans-type
      "EFTPOS DEBIT" (debit-account-to trans statement)
      "TRANSFER CREDIT" (transfer-credit-account-to trans statement)
      "MISCELLANEOUS DEBIT" (debit-account-to trans statement)
      "TRANSFER DEBIT" (transfer-debit-account-to trans statement)
      "INTEREST PAID" (interest-account-to trans statement)
      "AUTOMATIC DRAWING" (automatic-drawing-account-to trans statement)
      "FEES" (fees-account-to trans statement)
      "REVERSAL CREDIT" (automatic-drawing-account-to trans statement)
      "ATM DEBIT" (atm-debit-account-to trans statement)
      "DEPOSIT" (deposit-account-to trans statement)
      "INTER-BANK CREDIT" (bank-credit-account-to trans statement)
      "MISCELLANEOUS CREDIT" (misc-credit-account-to trans statement)
      nil))) ; this should really be handled better, for "types" we haven't accounted for (debit rate going up?!? 21 NOV 2022)

(defn determine-category-savings
  "The savings version of the above function (ugh)"
  [trans statement]
  (let [trans-type (transaction-type trans)]
    (case trans-type
      "TRANSFER CREDIT" (transfer-credit-savings trans statement)
      "TRANSFER DEBIT" (transfer-debit-savings trans statement)
      "INTEREST PAID" (interest-savings trans statement)
      "INTER-BANK CREDIT" (bank-credit-account-to trans statement)
      "DIVIDEND" (dividend-savings trans statement)
      "SALARY" (salary-savings trans statement))))

(defn write-statement
  [statement]
  (spit beancount-file (clojure.string/join "\n" (conj statement "\n")) :append true))

(defn tap
  [s]
  (println s)
  s)

(defn organize-correctly
  [statement]
  (-> statement
      (flatten)
      (into [])
      (->> (clojure.string/join " "))
      (->> (conj []))))

(def accounts (atom #{}))
(def new-accounts (atom #{}))

(defn create-account
  [transaction]
  (let [to-account (second (clojure.string/split (second transaction) #"  "))
        from-account (second (clojure.string/split (nth transaction 2) #"  "))
        marshalled-from-account (clojure.string/replace from-account #"! " "")]
    (if (not (contains? (deref accounts) to-account))
      (swap! new-accounts conj to-account))
    (if (not (contains? (deref accounts) marshalled-from-account))
      (swap! new-accounts conj marshalled-from-account))
    transaction))

(defn do-we-continue?
  "we need to do some fuckery because we're handling nil as the case where we don't want to write anything
  It's sorta an escape hatch, or the break point"
  [continue?]
  (if (not (nil? continue?))
    (->> continue?
         (create-account)
         (write-statement))))

(defn write-transaction-to-file
  "Please re-write this function so that make-date returns a map instead of a list,
  it should be a parsed data structure with the statement as one of the keys"
  [transaction]
  (->> []
       (make-date transaction)
       (add-transaction-flag)
       (add-narration transaction)
       (organize-correctly)
       (add-account-from-and-amount transaction)
       (determine-category transaction)
       (do-we-continue?)))

(defn write-transaction-to-file-savings
  [transaction]
  (->> []
       (make-date transaction)
       (add-transaction-flag)
       (add-narration transaction)
       (organize-correctly)
       (add-account-from-and-amount-savings transaction)
       (determine-category-savings transaction)
       (do-we-continue?)))

(defn index-of [x coll]
  (let [idx? (fn [i a] (when (= x a) i))]
    (first (keep-indexed idx? coll))))

(defn get-account
  "Returns the account from statements where the account is opened"
  [account-info]
  (let [split (into [] (clojure.string/split account-info #" "))
        account (last split)]
    account))

(defn read-accounts
  [ledger]
  (let [file-contents (slurp ledger)
        split (into [] (rest (clojure.string/split file-contents #"\n")))
        empty-index (index-of "" split)
        only-accounts (if (not (nil? empty-index))
                        (subvec split 0 empty-index)
                        split)]
    (->> (mapv get-account only-accounts)
         (set)
         (reset! accounts))))

(defn open-account-string
  [account]
  (str "2016-01-01 open " account))

(defn write-accounts
  " TODO
  this function doesn't look at @accounts..."
  [ledger]
  (let [file-contents (slurp ledger)
        split (clojure.string/split file-contents #"\n")
        empty-index (index-of "" split)
        [before after] (split-at empty-index split)
        middle (mapv open-account-string (into [] (deref new-accounts)))
        full-file (vec (concat before middle after "\n"))]
    (spit ledger (clojure.string/join "\n" full-file))))

(defn move-beancount-file
  []
  #_(sh "cp" beancount-file "/home/anish/usr/finance/")
  (sh "scp" beancount-file "mossnet.lan:~/usr/finance/ledger.beancount"))

(defn backup-beancount-file
  []
  (sh "cp" beancount-file "./documents/backup.beancount"))

(defn assert-balance
  [transaction checking?]
  (let [b (balance transaction)
        account (if checking?
                  "Assets:Checking"
                  "Assets:Savings")
        date (-> (sh "date" "+%Y-%m-%d")
                 :out
                 (string/trim-newline))]
    (write-statement (vector (str date " balance " account " " b " AUD")))))

(defn main
  [csv checking? & _args]
  (read-accounts beancount-file)
  (if checking?
    (mapv write-transaction-to-file (take-csv csv))
    (mapv write-transaction-to-file-savings (take-csv csv)))
  (write-accounts beancount-file)
  (assert-balance (first (take-csv csv)) checking?)
  (reset! new-accounts #{}))
